#!/usr/bin/perl
#
# Create some data we can use to test the "collate" and "photom" programs.
# Then run those programs to make sure that they are working properly.
#
# Return with code 0 if all goes well,
#                  1 if an error(s) occurs -- and print message
#
# MWR 6/9/2001
#
# Added tests of extinction.
#
# MWR 8/23/2002
#
# Added tests for the new "fixk" option.  One test should yield good results,
#   but the other test _should_ yield bad results.  That will cause some
#   warning messages to print to the screen; so we start by printing
#   a line alerting the user that some warnings are about to follow,
#   but don't panic.
#
# MWR 2/16/2003


use strict;
use POSIX;

my($debug);
my($nstar);
my($nimages);
my($i_image);
my(@coeff_a_v, $coeff_b_v);
my(@coeff_a_i, $coeff_b_i);
my($coeff_k_v, $coeff_k_i);
my(@list_stem);
my($catalog_file);
my($retcode);
my($mode);

# if this is set to 1, then we print some diagnostic messages during
#   execution and don't delete temporary files
$debug = 0;

# mode of photometric solution: 0 for no extinction, 1 for extinction
# 
# we run a loop here to test both cases ... not elegant, but it will
#   do the trick
for ($mode = 0; $mode <= 1; $mode++) {

  # number of stars
  $nstar = 200;
  # number of different pairs of images
  $nimages = 5;
  if ($mode == 0) {
	 # zero-point offset coefficients
	 for ($i_image = 0; $i_image < $nimages; $i_image++) {
		$coeff_a_v[$i_image] = 1.0 - $i_image*0.1;
		$coeff_a_i[$i_image] = 1.1 - $i_image*0.1;
	 }
	 # extinction coefficients
	 $coeff_k_v = 0.0;
	 $coeff_k_i = 0.0;
  } 
  else {
	 # zero-point offset coefficients
	 for ($i_image = 0; $i_image < $nimages; $i_image++) {
		$coeff_a_v[$i_image] = 1.0;
		$coeff_a_i[$i_image] = 1.1;
	 }
	 # extinction coefficients
	 $coeff_k_v = 0.20;
	 $coeff_k_i = 0.06;
  }
  # color-term coefficients
  $coeff_b_v = 0.05;
  $coeff_b_i = 0.10;
  # name of catalog file
  $catalog_file = "catalog.dat";
  # stem of file name for raw star lists
  for ($i_image = 0; $i_image < $nimages; $i_image++) {
	 $list_stem[$i_image] = "hxra202870$i_image.fits.ast";
  }

  $retcode = 0;

  # generate a set of artificial stars, and place lists of their properties
  #   into output file(s) which can be used as input to "collate"
  if (generate_stars($nstar, $nimages, @coeff_a_v, @coeff_a_i, 
								 $coeff_b_v, $coeff_b_i, $coeff_k_v, $coeff_k_i,
								 $catalog_file, @list_stem) != 0) {
	 die("selftest.pl: failed to generate artificial star data");
  }


  # run the "collate" program to combine the two lists of raw photometry
  #   into a single file which contains both V and I data
  if (run_collate($nimages, @list_stem) != 0) {
	 die("selftest.pl: collate failed to run properly");
  }


  # run the "photom" program to create a photometric solution for the 
  #   night, using the collated output.  
  if (run_photom($nimages, $mode, $catalog_file, @list_stem) != 0) {
	 die("selftest.pl: run_photom failed to run properly");
  }


  # check the results of the "photom" program
  if ($mode == 0) {
    if (check_results($nimages, @coeff_a_v, @coeff_a_i, 
                                  $coeff_b_v, $coeff_b_i) != 0) {
	   die("selftest.pl: check_results finds error in photometric solution");
    }
  } 
  elsif ($mode == 1) {
    if (check_results($nimages, @coeff_a_v, @coeff_a_i, 
                                  $coeff_b_v, $coeff_b_i,
                                  $coeff_k_v, $coeff_k_i) != 0) {
	   die("selftest.pl: check_results finds error in photometric solution");
    }
  } 


  # check the action of the "fixk" option.  If this is mode=0,
  #   so that the artificial data was generated with no extinction,
  #   we run photom twice:
  #
  #        first, with fixed extinction  kv=0.0, ki=0.0
  #                (this should give perfect results, RMS = 0.0)
  #
  #        next, with fixed extinction   kv=0.20, ki=0.08
  #                (this should give sizeable residuals, worse in V)
  # 
  # Verify that the RMS is zero in the first case, and verify that
  #   the residuals are NOT zero in the second case (in fact, 
  #   they should be 2.5 times bigger in V than in I).
  #
  if ($mode == 0) {
    # run the "photom" program to exercise the "fixk" option, 
    #   with extinction values fixed at 0.0
    if (run_photom($nimages, 2, $catalog_file, @list_stem) != 0) {
      die("selftest.pl: run_photom failed to run properly");
    }
    # results should be good this time, since there was no extinction
    #   put into the data
    if ($debug > 0) {
      printf "about to check results of fixk=0 -- should be GOOD \n";
    }
    if (check_results($nimages, @coeff_a_v, @coeff_a_i, 
                                  $coeff_b_v, $coeff_b_i) != 0) {
	   die("selftest.pl: check_results finds error in solution, fixk=0 case");
    }

    # Now run the "photom" program to exercise the "fixk" option, 
    #   with extinction values fixed at 0.20, 0.08 -- not input values
    if (run_photom($nimages, 3, $catalog_file, @list_stem) != 0) {
      die("selftest.pl: run_photom failed to run properly");
    }
    # results should be bad this time, since there was no extinction
    #   put into the data.  Warn the user that some error messages
    #   will appear.
    if ($debug > 0) {
      printf "about to check results of fixk=0.20 -- should be BAD \n";
    }
    printf "  a set of messages with zero-point mismatches should follow ...\n";
    if (check_results($nimages, @coeff_a_v, @coeff_a_i, 
                                  $coeff_b_v, $coeff_b_i) == 0) {
	   die("selftest.pl: check_results finds error in solution, fixk=0.20 case");
    }
  } 


  # clean up
  if (clean_up($nimages, $catalog_file, @list_stem) != 0) {
	 die("selftest.pl: failed to clean up properly");
  }

}

exit($retcode);



##############################################################################
# Generate data on artificial stars, and place it into raw star lists
#   which we can use as input to the "collate" program.
#
# We don't QUITE generate the magnitudes properly, so our color terms
#   are going to be a little bit different from the expected color terms --
#   but we should be in the ballpark.
#
# RETURNS:
#   0          if all goes well
#   1          if there's an error
#
sub generate_stars {
  my($nstar, $nimages);
  my(@coeff_a_v, @coeff_a_i, $coeff_b_v, $coeff_b_i, @list_stem);
  my($i_image);
  my($i, $id, $ra, $dec, $vmag, $imag, $color, $scatter);
  my($start_vmag, $start_imag);
  my($ra_min, $dec_min, $pos_spacing);
  my($mag_spacing, $vmag_sig, $imag_sig);
  my($output_file_v, $output_file_i);
  my($sky, $skysig, $quality);
  my($row, $col, $row_spacing);
  my($peak, $fwhm, $round, $sharp);
  my($catalog_v, $catalog_i, $catalog_file);
  my($this_v, $this_i);
  my($jd, $airmass, $exptime);
  my(@vmag_array, @imag_array);

  $nstar = $_[0];
  $nimages = $_[1];
  for ($i = 0; $i < $nimages; $i++) {
    $coeff_a_v[$i] = $_[2 + $i];
    $coeff_a_i[$i] = $_[2 + $nimages + $i];
    if ($debug > 0) {
      printf "generate_stars: coeff_a_v $i is $coeff_a_v[$i] \n";
      printf "generate_stars: coeff_a_i $i is $coeff_a_i[$i] \n";
    }
  }
  $coeff_b_v = $_[2 + 2*$nimages];
  $coeff_b_i = $_[3 + 2*$nimages];
  $coeff_k_v = $_[4 + 2*$nimages];
  $coeff_k_i = $_[5 + 2*$nimages];
  if ($debug > 0) {
    printf "generate_stars: coeff_b_v is $coeff_b_v \n";
    printf "generate_stars: coeff_b_i is $coeff_b_i \n";
    printf "generate_stars: coeff_k_v is $coeff_k_v \n";
    printf "generate_stars: coeff_k_i is $coeff_k_i \n";
  }
  $catalog_file = $_[6 + 2*$nimages];
  for ($i = 0; $i < $nimages; $i++) {
    $list_stem[$i] = $_[7 + 2*$nimages + $i];
  }


 for ($i_image = 0; $i_image < $nimages; $i_image++) {

  # open the four output files: 2 for .coo, 2 for .ast
  # first, the .ast files
  $output_file_v = $list_stem[$i_image];
  $output_file_v =~ s/x/v/;
  $output_file_i = $list_stem[$i_image];
  $output_file_i =~ s/x/i/;
  open(OUTPUT_V_AST, ">$output_file_v") || 
           die("generate_stars: can't open file $output_file_v for writing");
  open(OUTPUT_I_AST, ">$output_file_i") || 
           die("generate_stars: can't open file $output_file_i for writing");
  # now, the .coo files
  $output_file_v =~ s/.ast/.coo/;
  $output_file_i =~ s/.ast/.coo/;
  open(OUTPUT_V_COO, ">$output_file_v") || 
           die("generate_stars: can't open file $output_file_v for writing");
  open(OUTPUT_I_COO, ">$output_file_i") || 
           die("generate_stars: can't open file $output_file_i for writing");

  # open the catalog file
  if ($i_image == 0) {
    open(CATALOG, ">$catalog_file") ||
           die("generate_stars: can't open file $catalog_file for writing");
  }


  # initialize properties of the first star -- we'll build from it
  $pos_spacing = 10.0/3600.0;
  $start_vmag = 10.0;
  $start_imag = 10.0;
  $vmag = $start_vmag;
  $imag = $start_imag;
  $mag_spacing = 0.01;
  $sky = 1000.0;
  $skysig = 1.0;
  $vmag_sig = 0.01;
  $imag_sig = 0.01;
  $quality = 0;
  $row = 10;
  $col = 10;
  $row_spacing = 10;
  $peak = 1000.0;
  $fwhm = 3.0;
  $round = 0.0;
  $sharp = 0.5;

  ($jd, $airmass, $ra_min, $dec_min, $exptime) = image_info($i_image);
  # we use a single airmass for all stars in each image; subtracting 
  #   a bit helps fix it to be like stars near center of each image
  $airmass -= 0.02;
  $ra = $ra_min;
  $dec = $dec_min;

  for ($i = 0; $i < $nstar; $i++) {
    $id = $i + 1;

    $color = ($vmag - $imag);
    $this_v = $vmag - $coeff_a_v[$i_image] - $coeff_b_v*($color) + 
                                $coeff_k_v*$airmass;
    $this_i = $imag - $coeff_a_i[$i_image] - $coeff_b_i*($color) +
                                $coeff_k_i*$airmass;
       
    # print this star to the output files
    printf OUTPUT_V_AST "%5d %9.5f %9.5f  %4d %5.2f  %6.3f %6.3f  %2d\n",
            $id, $ra, $dec, $sky, $skysig, $this_v, $vmag_sig, $quality;
    printf OUTPUT_I_AST "%5d %9.5f %9.5f  %4d %5.2f  %6.3f %6.3f  %2d\n",
            $id, $ra, $dec, $sky, $skysig, $this_i, $imag_sig, $quality;

    printf OUTPUT_V_COO "%5d %7.2f %7.2f  %5d %6.3f  %7.3f %6.3f \n",
            $id, $row, $col, $peak, $fwhm, $round, $sharp;
    printf OUTPUT_I_COO "%5d %7.2f %7.2f  %5d %6.3f  %7.3f %6.3f \n",
            $id, $row, $col, $peak, $fwhm, $round, $sharp;

    # and, every now and then, put this star into the catalog file

    if (($i_image == 0) && ($i % 10 == 0)) {
      $catalog_v = $vmag;
      $catalog_i = $imag;
      printf CATALOG "%9.5f %9.5f  %6.3f  %6.3f \n", $ra, $dec, 
               $catalog_v, $catalog_i;
    } 

    # move onto the properties of the next star
    $ra += $pos_spacing;
    $dec += $pos_spacing;
    if ($i_image == 0) {
      $vmag_array[$i] = $start_vmag + (50.0 - rand(101))*$mag_spacing;
    }
    $vmag = $vmag_array[$i];
    if ($i_image == 0) {
      $imag_array[$i] = $start_imag + (50.0 - rand(101))*$mag_spacing;
    }
    $imag = $imag_array[$i];
    $row += $row_spacing;
    $col += $row_spacing;
  }

  if ($i_image == 0) {
    close(CATALOG);
  }

 }

  close(OUTPUT_V_AST);
  close(OUTPUT_I_AST);
  close(OUTPUT_V_COO);
  close(OUTPUT_I_COO);


  return(0);
}



#############################################################################
# Run the "collate" program on the test data files to produce a set
#   of merged data file, which may then be used as input for the
#   "photom" program.
#
# RETURNS
#    0          if all goes well
#    1          if there's an error
#
sub run_collate {
  my($nimages, $i_image);
  my(@list_stem);
  my($matchrad);
  my($jd, $airmass);
  my($ra, $dec);
  my($latitude);
  my($longitude);
  my($exptime);
  my($outfile);
  my($ast_file_v, $coo_file_v, $ast_file_i, $coo_file_i);
  my($pass_v, $pass_i);
  my($cmd_str);


  $nimages = $_[0];
  for ($i_image = 0; $i_image < $nimages; $i_image++) {
    $list_stem[$i_image] = $_[1 + $i_image];
  }

  # now, we generate one .clt file per pair of images
  for ($i_image = 0; $i_image < $nimages; $i_image++) {
  
    $outfile = get_collate_filename($list_stem[$i_image]);
    if ($debug > 0) {
      printf "run_collate: outfile will be ..$outfile..\n";
    }
    $pass_v = "V";
    $ast_file_v = $list_stem[$i_image];
    $ast_file_v =~ s/x/v/;
    $coo_file_v = get_coo_filename($ast_file_v);
    $pass_i = "I";
    $ast_file_i = $list_stem[$i_image];
    $ast_file_i =~ s/x/i/;
    $coo_file_i = get_coo_filename($ast_file_i);

    # match radius in arcsec
    $matchrad = 1.0;
    ($jd, $airmass, $ra, $dec, $exptime) = image_info($i_image);
    ($latitude, $longitude) = get_latitude_longitude();

    # run the collate program
    $cmd_str = "./collate passband=$pass_v $coo_file_v $ast_file_v ";
    $cmd_str = $cmd_str . " passband=$pass_i $coo_file_i $ast_file_i ";
    $cmd_str = $cmd_str . " matchrad=$matchrad jd=$jd ";
    $cmd_str = $cmd_str . " lat=$latitude long=$longitude ";
    $cmd_str = $cmd_str . " exptime=$exptime outfile=$outfile ";
    if ($debug > 0) {
      printf "run_collate: cmd_str is ..$cmd_str..\n";
    }
  
    $retcode = system($cmd_str);
    if ($debug > 0) {
      printf "retcode is ..$retcode..\n";
    }
    if ($retcode != 0) {
      warn "run_collate: the collate program fails on test data";
      return(1);
    }

  }

  return(0);
}


##############################################################################
# Convert the filename stem to the appropriate .coo file name.
#
# RETURNS
#   the .coo output file name
#
sub get_coo_filename {
  my($list_stem, $outfile);

  $list_stem = $_[0];

  $outfile = $list_stem;
  $outfile =~ s/.fits.ast/.fits.coo/;

  return($outfile);
}




##############################################################################
# Convert the filename stem to the appropriate collate output file name.
#
# RETURNS
#   the collated output file name
#
sub get_collate_filename {
  my($list_stem, $outfile);

  $list_stem = $_[0];

  $outfile = $list_stem;
  $outfile =~ s/hx/Mh/;
  $outfile =~ s/.fits.ast/.clt/;

  return($outfile);
}



#############################################################################
# Delete temporary files.
#
# RETURNS
#    0          if all goes well
#    1          if there's an error
#
sub clean_up {
  my($nimages, $i_image);
  my($catalog_file);
  my(@list_stem);
  my($file);

  $nimages = $_[0];
  $catalog_file = $_[1];
  for ($i_image = 0; $i_image < $nimages; $i_image++) {
    $list_stem[$i_image] = $_[2 + $i_image];
  }

  if ($debug == 1) {
    printf "clean_up: doing nothing, since debug = 1\n";
    return(0);
  }

  # delete the catalog file
  unlink($catalog_file);

  # delete the temporary .ast and .coo files
  for ($i_image = 0; $i_image < $nimages; $i_image++) {
    $file = $list_stem[$i_image];
    $file =~ s/x/v/;
    unlink($file);
    $file = get_coo_filename($file);
    unlink($file);

    $file = $list_stem[$i_image];
    $file =~ s/x/i/;
    unlink($file);
    $file = get_coo_filename($file);
    unlink($file);

    # delete the collated output file
    $file = get_collate_filename($list_stem[$i_image]);
    unlink($file);
  }

  # delete all the files created by the photom program
  $file = "photom.coeff";
  unlink($file);
  $file = "photom_V.comp";
  unlink($file);
  $file = "photom_I.comp";
  unlink($file);
  $file = "photom.cal";
  unlink($file);
  $file = "selftest.debug";
  unlink($file);

  return(0);
}


#############################################################################
# Run the "photom" program on collated output file(s);
#   if all goes well, this will produce a set of output files 
#   we'll check later
#
# RETURNS
#    0          if all goes well
#    1          if there's an error
#
sub run_photom {
  my($nimages, $i_image);
  my($catalog_file);
  my(@list_stem);
  my(@infile);
  my($cat_racol, $cat_deccol, $cat_nfilt, $cat_vcols, $cat_icols);
  my($det_racol, $det_deccol, $det_nfilt);
  my($det_jdcol, $det_aircol, $det_flagcol);
  my($det_vcols, $det_icols);
  my($matchrad);
  my($cmd_str);
  my($mode);


  $nimages = $_[0];
  # the 'mode' value means
  #     0        "noext", with no fixed values for ext coeffs
  #     1        "ext", solve for ext coeffs
  #     2        "noext", with fixed values kv=0.00, ki=0.00
  #     3        "noext", with fixed values kv=0.20, ki=0.08
  $mode = $_[1];
  $catalog_file = $_[2];
  for ($i_image = 0; $i_image < $nimages; $i_image++) {
    $list_stem[$i_image] = $_[3 + $i_image];
  }
  

  for ($i_image = 0; $i_image < $nimages; $i_image++) {
    $infile[$i_image] = get_collate_filename($list_stem[$i_image]);
    if ($debug > 0) {
      printf "run_photom: collate input file is ..$infile[$i_image]..\n";
    }
  }

  $cat_racol = 0;
  $cat_deccol = 1;
  $cat_nfilt = 2;
  $cat_vcols = "V=2";
  $cat_icols = "I=3";

  $det_racol = 1;
  $det_deccol = 2;
  $det_jdcol = 3;
  $det_aircol = 4;
  $det_flagcol = 11;
  $det_nfilt = 2;
  $det_vcols = "5,6,7";
  $det_icols = "8,9,10";

  # match radius in arcsec
  $matchrad = 1.0;

  # run the photom program
  $cmd_str = "./photom $catalog_file $cat_racol $cat_deccol";
  $cmd_str = $cmd_str . " $cat_nfilt $cat_vcols $cat_icols";
  $cmd_str = $cmd_str . " $infile[0] ";
  $cmd_str = $cmd_str . " $det_racol $det_deccol $det_jdcol ";
  $cmd_str = $cmd_str . " $det_aircol $det_flagcol ";
  $cmd_str = $cmd_str . " $det_nfilt $det_vcols $det_icols";
  for ($i_image = 1; $i_image < $nimages; $i_image++) {
    $cmd_str = $cmd_str . " $infile[$i_image] ";
  }
  if ($mode == 0) {
    $cmd_str .= " mode=noext ";
  }
  elsif ($mode == 1) {
    $cmd_str .= " mode=ext ";
  } 
  elsif ($mode == 2) {
    $cmd_str .= " mode=noext fixk=V,0.0 fixk=I,0.0 ";
  }
  elsif ($mode == 3) {
    $cmd_str .= " mode=noext fixk=V,0.20 fixk=I,0.08 ";
  }
  else {
    printf "run_photom: invalid mode value ..$mode..\n";
    exit(1);
  }
  if ($debug > 0) {
    $cmd_str = $cmd_str . " radius=$matchrad >& selftest.debug";
  }
  else {
    $cmd_str = $cmd_str . " radius=$matchrad >& /dev/null ";
  }
  if ($debug > 0) {
    printf "run_photom: cmd_str is ..$cmd_str..\n";
  }

  # 
  #  exit here to leave test data lying in directory
  #

  $retcode = system($cmd_str);
  if ($debug > 0) {
    printf "retcode is ..$retcode..\n";
  }
  if ($retcode != 0) {
    warn "run_photom: the photom program fails on test data";
    return(1);
  }

  return(0);
}


##############################################################################
# Check the results of the photometric solution on the test data.
#   We look only at the values in the "photom.coeff" file, ignoring
#   all the other files we _could_ check.
#
# We pass the expected solution values to this routine:
#   arg 0       # of zero-points to check 
#       1       zero-pt in V for first pair
#       2       zero-pt in V for second pair
#      ...
#       N       zero-pt in V for N'th pair
#      N+1      zero-pt in I for first pair
#      N+2      zero-pt in I for second pair
#      ...
#      2N       zero-pt in I for N'th pair
#     2N+1      color term for V (is same for all image pairs)
#     2N+2      color term for I (is same for all image pairs)
#            (and, possibly)
#     2N+3      extinction term for V (is same for all pairs)
#     2N+4      extinction term for V (is same for all pairs)
#
# RETURNS:
#    0            if all goes well
#    1            if there's an error
#
sub check_results {
  my($nimages, $i_image);
  my(@coeff_a_v, @coeff_a_i, $coeff_b_v, $coeff_b_i);
  my($coeff_k_v, $coeff_k_i);
  my($coeff_file);
  my($nlines);
  my($line, @words);
  my($solution_a, $solution_b, $solution_k, $solution_rms);
  my($retcode);

  $nimages = $_[0];
  for ($i_image = 0; $i_image < $nimages; $i_image++) {
    $coeff_a_v[$i_image] = $_[1 + $i_image];
    $coeff_a_i[$i_image] = $_[1 + $nimages + $i_image];
  }
  $coeff_b_v = $_[2*$nimages + 1];
  $coeff_b_i = $_[2*$nimages + 2];
  $coeff_k_v = $_[2*$nimages + 3];
  $coeff_k_i = $_[2*$nimages + 4];
  if ($debug > 0) {
    printf "check_results: expect solutions with \n";
	 for ($i_image = 0; $i_image < $nimages; $i_image++) {
	   printf "V: pair %2d  a %6.3f  b %6.3f ", 
                              $i_image, $coeff_a_v[$i_image], $coeff_b_v;
      if ($mode == 1) {
        printf "  k %6.2f ", $coeff_k_v;
      }
      printf "\n";

	   printf "I: pair %2d  a %6.3f  b %6.3f ", 
                              $i_image, $coeff_a_i[$i_image], 0.0-$coeff_b_i;
      if ($mode == 1) {
        printf "  k %6.2f ", $coeff_k_i;
      }
      printf "\n";

    }
  }

  $retcode = 0;

  $coeff_file = "photom.coeff";

  # we read the results from the "photom.coeff" file
  if (open(COEFF, $coeff_file) == undef) {
    warn("check_results: can't open file $coeff_file with photometric results");
    return(1);
  }

  # there should be two lines in the COEFF file for each image:
  #   one for solution in V, the other for solution in I. 
  #   The lines are arranged like so:
  #        V sol for image pair 0
  #        V sol for image pair 1
  #           ....
  #        I sol for image pair 0
  #        I sol for image pair 1
  #           ....
  $nlines = 0;
  while (<COEFF>) {
    $line = $_;
    @words = split(/\s+/, $line);
    $solution_a = $words[5];
    $solution_b = $words[8];
    if ($mode == 0) {
      $solution_rms = $words[11];
    }
    elsif ($mode == 1) {
      # note that the value of "k" is given as negative in output .coeff
      $solution_k = 0.0 - $words[11];
      $solution_rms = $words[14];
    }
  
    if ($debug > 0) {
      printf "soln %3d: a %6.3f b %6.3f ", 
                 $nlines, $solution_a, $solution_b;
      if ($mode == 1) {
        printf " k %6.3f ", $solution_k;
      }
      printf " RMS %6.3f \n", $solution_rms;
    }

    # Check results for V
    if ($mode == 0) {
		if ($nlines < $nimages) {
		  $i_image = $nlines;
		  if (are_the_same($coeff_a_v[$i_image], $solution_a) == 0) {
			 printf "check_results: pair %2d V-band zero point is %6.3f   should be %6.3f \n",
								$i_image, $solution_a, $coeff_a_v[$i_image];
			 $retcode = 1;
		  }
		  if (are_the_same($coeff_b_v, $solution_b) == 0) {
			 printf "check_results: pair %2d V-band color term is %6.3f   should be %6.3f \n",
								$i_image, $solution_b, $coeff_b_v;
			 $retcode = 1;
		  }
		  if ($mode == 1) {
			 if (are_the_same($coeff_k_v, $solution_k) == 0) {
				printf "check_results: pair %2d V-band extinction term is %6.3f   should be %6.3f \n",
								$i_image, $solution_k, $coeff_k_v;
				$retcode = 1;
			 }
		  }
		}
    }
    elsif ($mode == 1) {
		if ($nlines == 0) {
		  $i_image = 0;
		  if (are_the_same($coeff_a_v[$i_image], $solution_a) == 0) {
			 printf "check_results: pair %2d V-band zero point is %6.3f   should be %6.3f \n",
								$i_image, $solution_a, $coeff_a_v[$i_image];
			 $retcode = 1;
		  }
		  if (are_the_same($coeff_b_v, $solution_b) == 0) {
			 printf "check_results: pair %2d V-band color term is %6.3f   should be %6.3f \n",
								$i_image, $solution_b, $coeff_b_v;
			 $retcode = 1;
		  }
		  if ($mode == 1) {
			 if (are_the_same($coeff_k_v, $solution_k) == 0) {
				printf "check_results: pair %2d V-band extinction term is %6.3f   should be %6.3f \n",
								$i_image, $solution_k, $coeff_k_v;
				$retcode = 1;
			 }
		  }
		}
    }

    # Check results for I
    if ($mode == 0) {
		if ($nlines >= $nimages) {
		  $i_image = $nlines - $nimages;
		  if (are_the_same($coeff_a_i[$i_image], $solution_a) == 0) {
			 printf "check_results: pair %2d I-band zero point is %6.3f   should be %6.3f \n",
								$i_image, $solution_a, $coeff_a_i[$i_image];
			 $retcode = 1;
		  }
		  # Note that due to the way we calculate colors, the I-band solution
		  #   for color term should be equal to the NEGATIVE of the input
		  #   coeff_b_i
		  $solution_b = 0 - $solution_b;
		  if (are_the_same($coeff_b_i, $solution_b) == 0) {
			 printf "check_results: pair %2d I-band color term is %6.3f   should be %6.3f \n",
								$i_image, $solution_b, $coeff_b_i;
			 $retcode = 1;
		  }
		  if ($mode == 1) {
			 if (are_the_same($coeff_k_i, $solution_k) == 0) {
				printf "check_results: pair %2d I-band extinction term is %6.3f   should be %6.3f \n",
								$i_image, $solution_k, $coeff_k_i;
				$retcode = 1;
			 }
		  }
		}
    }
    elsif ($mode == 1) {
		if ($nlines == 1) {
		  $i_image = 0;
		  if (are_the_same($coeff_a_i[$i_image], $solution_a) == 0) {
			 printf "check_results: pair %2d I-band zero point is %6.3f   should be %6.3f \n",
								$i_image, $solution_a, $coeff_a_i[$i_image];
			 $retcode = 1;
		  }
		  # Note that due to the way we calculate colors, the I-band solution
		  #   for color term should be equal to the NEGATIVE of the input
		  #   coeff_b_i
		  $solution_b = 0 - $solution_b;
		  if (are_the_same($coeff_b_i, $solution_b) == 0) {
			 printf "check_results: pair %2d I-band color term is %6.3f   should be %6.3f \n",
								$i_image, $solution_b, $coeff_b_i;
			 $retcode = 1;
		  }
		  if ($mode == 1) {
			 if (are_the_same($coeff_k_i, $solution_k) == 0) {
				printf "check_results: pair %2d I-band extinction term is %6.3f   should be %6.3f \n",
								$i_image, $solution_k, $coeff_k_i;
				$retcode = 1;
			 }
		  }
		}
    }

    # This should never happen
    if ($nlines >= 2*$nimages) {
      warn "check_results: too many lines in file $coeff_file?";
      $retcode = 1;
    }

    $nlines++;
  }

  return($retcode);
}


##############################################################################
# Compare two numbers, and decide if they are "close enough" 
#   to denote a successful execution.
#
# RETURNS
#     0                if the two are NOT the same
#     1                if the two are close enough
#
sub are_the_same {
  my($x, $y, $diff, $acceptable_diff);

  $x = $_[0];
  $y = $_[1];

  $acceptable_diff = 0.10;

  $diff = $x - $y;
  if (abs($diff) > $acceptable_diff) {
    return(0);
  }
  else {
    return(1);
  }

}



##############################################################################
# PROCEDURE: image_info
#
# DESCRIPTION: Given an image index number, 
#              calculate the JD and the airmass of the image (at midexposure).
#              Also supply a fixed RA, Dec, and exposure time (in seconds).
#
# RETURNS
#              JD, airmass, RA, Dec, exptime   as five-component list
#
sub image_info {
  my($ra, $dec, $image_index);
  my($jdbase, $latitude, $longitude, $exptime);
  my($jd, $airmass, $exptime);
  my($degtorad);
  my($long, $j, $s, $t, $t0, $lst);
  my($ha);
  my($x, $u, $z);
  my($alt);
  my($a, $sl, $cl, $sh, $ch, $sd, $cd, $y, $z, $xx);

  $degtorad = (3.1415926/180.0);

  $image_index = $_[0];
  if ($debug > 0) {
    printf "image_info:  index %3d \n", $image_index;
  }

  $ra = 180.0;
  $dec = 1.0;
  $jdbase = 2452029.700;
  $latitude = 45.0;
  $longitude = 90.0;
  $exptime = 1.0;

  $jd = $jdbase + $image_index*0.01;
  if ($debug > 0) {
    printf "image_info: JD is %12.5f \n", $jd;
  }

  # first, get the LST corresponding to this JD at this location
  #   convert longitude from decimal degrees to hours West of Greenwich
  $long = $longitude/15.0;
  $j = floor($jd + 0.5) - 0.5;
  $s = $j - 2451545.0;
  $t = $s/36525.0;
  $t0 = 6.697374558 + 2400.051336*$t + 0.000025862*$t*$t;
  while ($t0 > 24.0) {
    $t0 -= 24.0;
  }
  while ($t0 < 0.0) {
      $t0 += 24.0;
  }
  $u = ($jd - $j)*24.0;
  $u *= 1.002737909;
  $t0 += $u;
  while ($t0 > 24.0) {
      $t0 -= 24.0;
  }
  while ($t0 < 0.0) {
      $t0 += 24.0;
  }
  $t0 -= $long;
  while ($t0 > 24.0) {
      $t0 -= 24.0;
  }
  while ($t0 < 0.0) {
      $t0 += 24.0;
  }
  $lst = $t0;
  
  if ($debug > 0) {
    printf "image_info: lst is $lst \n";
  }

  # now, find the Hour Angle of the object, and convert to decimal degrees
  $x = 15.0*$lst;
  if ($x == 360.0) {
    $x = 0.0;
  }
  $ha = $x - $ra;
  if ($debug > 0) {
    printf "image_info: HA is $ha \n";
  }

  # next, calculate the airmass 
  $sl = sin($degtorad*$latitude);
  $cl = cos($degtorad*$latitude);
  $sh = sin($degtorad*$ha);
  $ch = cos($degtorad*$ha);
  $sd = sin($degtorad*$dec);
  $cd = cos($degtorad*$dec);

  $a = $sl*$sd + $cl*$cd*$ch;
  $alt = 90.0 - (acos($a)/$degtorad);
  if ($debug > 0) {
    printf "image_info: altitude is %5.1f degrees \n", $alt;
  }
  if ($alt < 1.0) {
    printf STDERR "image_info: bad altitude %7.1f degrees?!\n";
    $airmass = 99.0;
  }
  else {
    $z = 1.0/$a;
    $xx = $z - 0.0018167*($z - 1.0) - 0.002875*($z - 1.0)*($z - 1.0)
                  - 0.008083*($z - 1.0)*($z - 1.0)*($z - 1.0);
    $airmass = $xx;
  }

  if ($debug > 0) {
    printf "image_info: JD %12.5f  airmass %6.2f  exptime %6.1f \n", 
                      $jd, $airmass, $exptime;
  }

  return($jd, $airmass, $ra, $dec, $exptime);

}


#############################################################################
# PROCEDURE: get_latitude_longitude
#
# DESCRIPTION: return the latitude and longitude, in decimal degrees.
#
# RETURNS:
#      lat and long in a 2-element list
#
sub get_latitude_longitude {
  my($latitude, $longitude);

  $latitude = 45.0;
  $longitude = 90.0;

  return($latitude, $longitude);
}
